#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "hash.h"
#include "pnclient.h"
#include "pnlog.h"

#ifndef FAIL
#define	FAIL	(-1)
#endif

#ifndef SUCC
#define SUCC	0
#endif

#ifndef ON
#define ON		1
#endif
#ifndef OFF
#define OFF		0
#endif

static hash_t *h_client=NULL;
static fd_set rfds;
static int maxfd=0;

static int pnclient_hash_comp_key(const void *key1, const void *key2)
{
    return *(int *)key1 == *(int *)key2 ? 0 : 1;
}

static int regist_completed(pnclient_t *cli)
{
	return cli->regist_complete_flag;
}

static void registration_complete(pnclient_t *cli)
{
	cli->regist_complete_flag = ON;
}


int pnclient_count(void)
{
	return hash_count(h_client);
}

int pnclient_init(void)
{
	h_client = hash_create(HASHCOUNT_T_MAX, pnclient_hash_comp_key, 0);
	if(!h_client) {
		return FAIL;
	}

	FD_ZERO(&rfds);

	return SUCC;
}


pnclient_t *pnclient_create(int sockfd, struct sockaddr_in cliaddr) 
{
	pnclient_t *cli;

	cli = calloc(sizeof(pnclient_t), 1);
	if(!cli) {
		pnlog_error("%s:%d:calloc error", __FILE__, __LINE__);
		return NULL;
	}

	cli->sockfd = sockfd;
	cli->cliaddr = cliaddr;
	cli->timestamp = time(NULL);
	cli->registration_complete = registration_complete;
	cli->regist_completed = regist_completed;
	cli->regist_complete_flag = OFF;

	return cli;
}

time_t pnclient_get_timestamp(pnclient_t *cli)
{
	return cli->timestamp;
}

void pnclient_update_timestamp(pnclient_t *cli, time_t unix_secs)
{
	if(unix_secs < cli->timestamp) {
		pnlog_warn("%s:%d:client_update_timestamp:out of date value", \
						__FILE__, __LINE__);
	}

	cli->timestamp = unix_secs;
}

fd_set pnclient_get_fdset(void)
{
	return rfds;
}

int pnclient_get_maxfd(void)
{
	return maxfd;
}


int pnclient_add(pnclient_t *add_cli) 
{
	int ret;

	if(!h_client) {
		pnlog_error("%s:%d:MUST do client_init first.", __FILE__, __LINE__);
		return FAIL;
	}
	ret = hash_alloc_insert(h_client, &add_cli->sockfd, add_cli);

	if( ret == FAIL ) return FAIL;

	/* add fd*/
	FD_SET(add_cli->sockfd, &rfds);
	if(add_cli->sockfd > maxfd) maxfd = add_cli->sockfd;

	return SUCC;
}

int pnclient_delete(pnclient_t *del_cli) 
{
	hnode_t *hn;

	hn = hash_lookup(h_client, &del_cli->sockfd);
	if(!hn) {
		pnlog_error("%s:%d:hash_lookup error", __FILE__, __LINE__);
		return FAIL;
	}

	/* hash free */
	hash_scan_delfree(h_client, hn);
	
	/* clear fd */
	FD_CLR(del_cli->sockfd, &rfds);
	/* close sockfd */
	pnsock_close(del_cli->sockfd);

	pnlog_info("client deleted %s:%d", inet_ntoa(del_cli->cliaddr.sin_addr), ntohs(del_cli->cliaddr.sin_port));

	if(del_cli) free(del_cli);

	return SUCC;
}

pnclient_t *pnclient_lookup(int sockfd)
{
	hnode_t *hn;

	hn = hash_lookup(h_client, &sockfd);

	return (pnclient_t *)hnode_get(hn);
}

hash_t *pnclient_get_hash(void)
{
	return h_client;
}


void pnclient_free(void)
{
	hash_allfree(h_client);
}
